from __future__ import annotations

import datetime as dt
from collections.abc import Callable
from math import isnan
from operator import eq
from typing import Any

import polars as pl
import sqlalchemy
from hypothesis import given
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    booleans,
    data,
    dates,
    datetimes,
    floats,
    integers,
    just,
    lists,
    none,
)
from polars import DataFrame, Datetime, Float64, Int64, PolarsDataType, Utf8
from polars.testing import assert_frame_equal
from pytest import mark, param, raises
from sqlalchemy import (
    Column,
    DateTime,
    Engine,
    Float,
    Integer,
    LargeBinary,
    MetaData,
    String,
    Table,
    select,
)

from utilities.datetime import UTC
from utilities.hypothesis import sqlite_engines, text_ascii
from utilities.polars import check_dataframe
from utilities.sqlalchemy import (
    InsertPolarsDataFrameError,
    TableColumnToExprError,
    ensure_tables_created,
    insert_polars_dataframe,
    select_to_polars_dataframe,
    table_column_to_expr,
)


def _check_datetime(x: dt.datetime | None, y: dt.datetime | None, /) -> bool:
    return (x == y) or (
        isinstance(x, dt.datetime)
        and isinstance(y, dt.datetime)
        and (x.replace(tzinfo=None) == y.replace(tzinfo=None))
    )


def _check_float(x: float | None, y: float | None, /) -> bool:
    return (
        (x == y)
        or ((x is None) and isinstance(y, float) and isnan(y))
        or (isinstance(x, float) and isnan(x) and (y is None))
    )


class TestInsertPolarsDataFrame:
    @given(data=data(), engine=sqlite_engines())
    @mark.parametrize(
        ("strategy", "pl_dtype", "col_type", "check"),
        [
            param(booleans() | none(), pl.Boolean, sqlalchemy.Boolean, eq),
            param(dates() | none(), pl.Date, sqlalchemy.Date, eq),
            param(datetimes() | none(), Datetime, DateTime, eq),
            param(
                datetimes(timezones=just(UTC)) | none(),
                Datetime(time_zone=UTC),
                DateTime(timezone=True),
                _check_datetime,
            ),
            param(floats(), Float64, Float, _check_float),
            param(integers(-10, 10) | none(), Int64, Integer, eq),
            param(text_ascii() | none(), Utf8, String, eq),
        ],
    )
    def test_main(
        self,
        *,
        data: DataObject,
        engine: Engine,
        strategy: SearchStrategy[Any],
        pl_dtype: PolarsDataType,
        col_type: Any,
        check: Callable[[Any, Any], bool],
    ) -> None:
        values = data.draw(lists(strategy))
        df = DataFrame({"value": values}, schema={"value": pl_dtype})
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("value", col_type),
        )
        ensure_tables_created(engine, table)
        insert_polars_dataframe(engine, df, table)
        sel = select(table.c["value"])
        with engine.begin() as conn:
            res = conn.execute(sel).scalars().all()
        for r, v in zip(res, values, strict=True):
            assert check(r, v)

    @given(engine=sqlite_engines(), values=lists(booleans() | none()))
    @mark.parametrize("sr_name", [param("Value"), param("value")])
    def test_snake(
        self, *, engine: Engine, values: list[bool | None], sr_name: str
    ) -> None:
        df = DataFrame({sr_name: values}, schema={sr_name: pl.Boolean})
        table = Table(
            "example",
            MetaData(),
            Column("Id", Integer, primary_key=True),
            Column("Value", sqlalchemy.Boolean),
        )
        ensure_tables_created(engine, table)
        insert_polars_dataframe(engine, df, table, snake=True)
        sel = select(table.c["Value"])
        with engine.begin() as conn:
            res = conn.execute(sel).scalars().all()
        assert res == values

    @given(engine=sqlite_engines(), values=lists(booleans() | none(), min_size=1))
    def test_polars_data_frame_yields_no_rows_error(
        self, *, engine: Engine, values: list[bool | None]
    ) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("value", sqlalchemy.Boolean),
        )
        df = DataFrame({"other": values}, schema={"other": pl.Boolean})
        with raises(InsertPolarsDataFrameError):
            insert_polars_dataframe(engine, df, table)


class TestSelectToPolarsDataFrame:
    @given(data=data(), engine=sqlite_engines())
    @mark.parametrize(
        ("strategy", "pl_dtype", "col_type"),
        [
            param(booleans() | none(), pl.Boolean, sqlalchemy.Boolean),
            param(dates() | none(), pl.Date, sqlalchemy.Date),
            param(datetimes() | none(), Datetime, DateTime),
            param(
                datetimes(timezones=just(UTC)) | none(),
                Datetime(time_zone=UTC),
                DateTime(timezone=True),
            ),
            param(floats(), Float64, Float),
            param(integers(-10, 10) | none(), Int64, Integer),
            param(text_ascii() | none(), Utf8, String),
        ],
    )
    def test_main(
        self,
        *,
        data: DataObject,
        engine: Engine,
        strategy: SearchStrategy[Any],
        pl_dtype: PolarsDataType,
        col_type: Any,
    ) -> None:
        values = data.draw(lists(strategy, min_size=1))
        df = DataFrame({"value": values}, schema={"value": pl_dtype})
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("value", col_type),
        )
        ensure_tables_created(engine, table)
        insert_polars_dataframe(engine, df, table)
        sel = select(table.c["value"])
        result = select_to_polars_dataframe(sel, engine)
        assert_frame_equal(result, df)

    @given(
        engine=sqlite_engines(),
        values=lists(booleans() | none()),
        batch_size=integers(1, 10),
    )
    def test_iter_batches(
        self, *, engine: Engine, values: list[bool | None], batch_size: int
    ) -> None:
        df = DataFrame({"value": values}, schema={"value": pl.Boolean})
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("value", sqlalchemy.Boolean),
        )
        ensure_tables_created(engine, table)
        insert_polars_dataframe(engine, df, table)
        sel = select(table.c["value"])
        dfs = select_to_polars_dataframe(
            sel, engine, iter_batches=True, batch_size=batch_size
        )
        for df in dfs:
            check_dataframe(
                df, min_height=1, max_height=batch_size, schema={"value": pl.Boolean}
            )

    @given(engine=sqlite_engines(), values=lists(booleans() | none()))
    def test_snake(self, *, engine: Engine, values: list[bool | None]) -> None:
        df = DataFrame({"Value": values}, schema={"Value": pl.Boolean})
        table = Table(
            "example",
            MetaData(),
            Column("Id", Integer, primary_key=True),
            Column("Value", sqlalchemy.Boolean),
        )
        ensure_tables_created(engine, table)
        insert_polars_dataframe(engine, df, table)
        sel = select(table.c["Value"])
        res = select_to_polars_dataframe(sel, engine, snake=True)
        expected = DataFrame({"value": values}, schema={"value": pl.Boolean})
        assert_frame_equal(res, expected)


class TestTableColumnToExpr:
    def test_error(self) -> None:
        column = Column("value", LargeBinary)
        with raises(TableColumnToExprError):
            _ = table_column_to_expr(column)



from __future__ import annotations

from collections.abc import Callable
from decimal import Decimal
from json import dumps
from math import isnan
from operator import neg
from typing import Any

from hypothesis import given
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    binary,
    booleans,
    characters,
    complex_numbers,
    data,
    dates,
    datetimes,
    decimals,
    dictionaries,
    floats,
    fractions,
    frozensets,
    integers,
    ip_addresses,
    just,
    lists,
    none,
    sets,
    slices,
    text,
    timedeltas,
    times,
    tuples,
    uuids,
)
from pytest import mark, param, raises
from typing_extensions import override

from utilities.datetime import NOW_HKG, UTC
from utilities.hypothesis import assume_does_not_raise, temp_paths, text_ascii
from utilities.json import (
    _CLASS,
    _VALUE,
    JsonDeserializationError,
    JsonSerializationError,
    deserialize,
    serialize,
)
from utilities.sentinel import sentinel


class TestSerialize:
    @given(data=data())
    @mark.parametrize(
        "strategy",
        [
            param(booleans()),
            param(characters()),
            param(dates()),
            param(datetimes(timezones=just(UTC) | none())),
            param(fractions()),
            param(ip_addresses()),
            param(lists(integers(), max_size=3)),
            param(none()),
            param(temp_paths()),
            param(text()),
            param(timedeltas()),
            param(times()),
            param(uuids()),
        ],
    )
    def test_main(self, *, data: DataObject, strategy: SearchStrategy[Any]) -> None:
        x = data.draw(strategy)
        ser_x = serialize(x)
        assert deserialize(ser_x) == x
        y = data.draw(strategy)
        res = ser_x == serialize(y)
        expected = x == y
        assert res is expected

    @given(x=binary(), y=binary())
    def test_binary(self, *, x: bytes, y: bytes) -> None:
        with assume_does_not_raise(UnicodeDecodeError):
            ser_x = serialize(x)
        assert deserialize(ser_x) == x
        with assume_does_not_raise(UnicodeDecodeError):
            res = ser_x == serialize(y)
        expected = x == y
        assert res is expected

    @given(x=complex_numbers(), y=complex_numbers())
    def test_complex(self, *, x: complex, y: complex) -> None:
        ser_x = serialize(x)

        def eq(x: complex, y: complex, /) -> bool:
            return ((x.real == y.real) or (isnan(x.real) and isnan(y.real))) and (
                (x.imag == y.imag) or (isnan(x.imag) and isnan(y.imag))
            )

        assert eq(deserialize(ser_x), x)
        res = ser_x == serialize(y)
        expected = eq(x, y)
        assert res is expected

    @given(x=decimals(), y=decimals())
    def test_decimal(self, *, x: Decimal, y: Decimal) -> None:
        ser_x = serialize(x)

        def eq(x: Decimal, y: Decimal, /) -> bool:
            x_nan, y_nan = x.is_nan(), y.is_nan()
            if x_nan and y_nan:
                return (x.is_qnan() == y.is_qnan()) and (x.is_signed() == y.is_signed())
            return (x_nan == y_nan) and (x == y)

        assert eq(deserialize(ser_x), x)
        res = ser_x == serialize(y)
        expected = eq(x, y)
        assert res is expected

    @given(data=data(), n=integers(0, 10))
    def test_dicts_sortable(self, *, data: DataObject, n: int) -> None:
        elements = dictionaries(
            integers(0, 2 * n), integers(0, 2 * n), min_size=n, max_size=n
        )
        x = data.draw(elements)
        ser_x = serialize(x)
        assert deserialize(ser_x) == x
        y = data.draw(elements)
        res = ser_x == serialize(y)
        expected = x == y
        assert res is expected

    @given(data=data(), n=integers(2, 10))
    def test_dicts_unsortable(self, *, data: DataObject, n: int) -> None:
        elements = dictionaries(
            integers(0, 2 * n) | text_ascii(min_size=1, max_size=1),
            integers(0, 2 * n),
            min_size=n,
            max_size=n,
        )
        x, y = data.draw(tuples(elements, elements))
        ser_x, ser_y = map(serialize, [x, y])
        if ser_x == ser_y:
            assert x == y

    @given(x=floats(), y=floats())
    def test_floats(self, *, x: float, y: float) -> None:
        ser_x = serialize(x)

        def eq(x: float, y: float, /) -> bool:
            return (x == y) or (isnan(x) and isnan(y))

        assert eq(deserialize(ser_x), x)
        res = ser_x == serialize(y)
        expected = eq(x, y)
        assert res is expected

    @given(data=data(), n=integers(0, 3))
    def test_slices(self, *, data: DataObject, n: int) -> None:
        x = data.draw(slices(n))
        ser_x = serialize(x)
        assert deserialize(ser_x) == x
        y = data.draw(slices(n))
        res = ser_x == serialize(y)
        expected = x == y
        assert res is expected

    @given(data=data(), n=integers(0, 10))
    @mark.parametrize("strategy", [param(frozensets), param(sets)])
    def test_sets_sortable(
        self, *, data: DataObject, n: int, strategy: Callable[..., SearchStrategy[int]]
    ) -> None:
        elements = strategy(integers(0, 2 * n), min_size=n, max_size=n)
        x = data.draw(elements)
        ser_x = serialize(x)
        assert deserialize(ser_x) == x
        y = data.draw(elements)
        res = ser_x == serialize(y)
        expected = x == y
        assert res is expected

    @given(data=data(), n=integers(2, 10))
    @mark.parametrize("strategy", [param(frozensets), param(sets)])
    def test_sets_unsortable(
        self,
        *,
        data: DataObject,
        n: int,
        strategy: Callable[..., SearchStrategy[int | str]],
    ) -> None:
        elements = strategy(
            integers(0, 2 * n) | text_ascii(min_size=1, max_size=1),
            min_size=n,
            max_size=n,
        )
        x, y = data.draw(tuples(elements, elements))
        ser_x, ser_y = map(serialize, [x, y])
        if ser_x == ser_y:
            assert x == y

    @given(data=data(), n=integers(0, 3))
    def test_tuples(self, *, data: DataObject, n: int) -> None:
        elements = tuples(*(n * [integers(0, 2 * n)]))
        x = data.draw(elements)
        ser_x = serialize(x)
        assert deserialize(ser_x) == x
        y = data.draw(elements)
        res = ser_x == serialize(y)
        expected = x == y
        assert res is expected

    @mark.parametrize(
        "obj",
        [param(NOW_HKG, id="invalid timezone"), param(sentinel, id="invalid object")],
    )
    def test_error(self, *, obj: Any) -> None:
        with raises(JsonSerializationError):
            _ = serialize(obj)

    @given(obj=dictionaries(text_ascii(), integers()))
    def test_deserializing_regular_dictionary(self, *, obj: dict[str, int]) -> None:
        ser = dumps(obj)
        assert deserialize(ser) == obj

    def test_deserialization_error(self) -> None:
        ser = dumps({_CLASS: "unknown", _VALUE: None})
        with raises(JsonDeserializationError):
            _ = deserialize(ser)

    @given(m=integers(), n=integers())
    def test_extra(self, *, m: int, n: int) -> None:
        class Example:
            def __init__(self, n: int, /) -> None:
                super().__init__()
                self.n = n

            @override
            def __eq__(self, other: object) -> bool:
                return isinstance(other, Example) and (self.n == other.n)

        def f_ser(obj: Example, /) -> int:
            return obj.n

        extra_ser = {Example: ("example", f_ser)}

        def f_des(n: int, /) -> Example:
            return Example(n)

        extra_des = {"example": f_des}
        x = Example(m)
        ser_x = serialize(x, extra=extra_ser)
        assert deserialize(ser_x, extra=extra_des) == x
        y = Example(n)
        res = ser_x == serialize(y, extra=extra_ser)
        expected = x == y
        assert res is expected

    def test_serialization_extra_error(self) -> None:
        class Example1:
            pass

        x = Example1()

        class Example2:
            pass

        extra = {Example2: ("example", neg)}
        with raises(JsonSerializationError):
            _ = serialize(x, extra=extra)

    @given(n=integers())
    def test_deserialization_extra_error(self, *, n: int) -> None:
        class Example:
            def __init__(self, n: int, /) -> None:
                super().__init__()
                self.n = n

        def f_ser(obj: Example, /) -> int:
            return obj.n

        x = Example(n)
        extra_ser = {Example: ("example", f_ser)}
        ser = serialize(x, extra=extra_ser)
        extra_des = {"wrong": neg}
        with raises(JsonDeserializationError):
            _ = deserialize(ser, extra=extra_des)



from __future__ import annotations

from pathlib import Path
from re import search
from time import sleep

from utilities.pyinstrument import profile


class TestProfile:
    def test_main(self, tmp_path: Path) -> None:
        with profile(path=tmp_path):
            sleep(1e-3)

        (file,) = tmp_path.iterdir()
        assert search(r"^profile__\d{8}T\d{6}\.html$", file.name)



from __future__ import annotations

import datetime as dt
from contextlib import suppress
from enum import Enum
from typing import Any, Generic, Literal, TypeVar

import luigi
from luigi import Parameter
from typing_extensions import override

from utilities.datetime import (
    EPOCH_UTC,
    ensure_date,
    ensure_datetime,
    ensure_time,
    parse_date,
    parse_datetime,
    parse_time,
    round_to_next_weekday,
    round_to_prev_weekday,
    serialize_date,
    serialize_datetime,
    serialize_time,
)
from utilities.enum import ensure_enum, parse_enum


class DateParameter(luigi.DateParameter):
    """A parameter which takes the value of a `dt.date`."""

    @override
    def normalize(self, value: dt.date | str) -> dt.date:
        return ensure_date(value)

    @override
    def parse(self, s: str) -> dt.date:
        return parse_date(s)

    @override
    def serialize(self, dt: dt.date) -> str:
        return serialize_date(dt)


class DateHourParameter(luigi.DateHourParameter):
    """A parameter which takes the value of an hourly `dt.datetime`."""

    def __init__(self, interval: int = 1, **kwargs: Any) -> None:
        super().__init__(interval, EPOCH_UTC, **kwargs)

    @override
    def normalize(self, dt: dt.datetime | str) -> dt.datetime:
        return ensure_datetime(dt)

    @override
    def parse(self, s: str) -> dt.datetime:
        return parse_datetime(s)

    @override
    def serialize(self, dt: dt.datetime) -> str:
        return serialize_datetime(dt)


class DateMinuteParameter(luigi.DateMinuteParameter):
    """A parameter which takes the value of a minutely `dt.datetime`."""

    def __init__(self, interval: int = 1, **kwargs: Any) -> None:
        super().__init__(interval=interval, start=EPOCH_UTC, **kwargs)

    @override
    def normalize(self, dt: dt.datetime | str) -> dt.datetime:
        return ensure_datetime(dt)

    @override
    def parse(self, s: str) -> dt.datetime:
        return parse_datetime(s)

    @override
    def serialize(self, dt: dt.datetime) -> str:
        return serialize_datetime(dt)


class DateSecondParameter(luigi.DateSecondParameter):
    """A parameter which takes the value of a secondly `dt.datetime`."""

    def __init__(self, interval: int = 1, **kwargs: Any) -> None:
        super().__init__(interval, EPOCH_UTC, **kwargs)

    @override
    def normalize(self, dt: dt.datetime | str) -> dt.datetime:
        return ensure_datetime(dt)

    @override
    def parse(self, s: str) -> dt.datetime:
        return parse_datetime(s)

    @override
    def serialize(self, dt: dt.datetime) -> str:
        return serialize_datetime(dt)


_E = TypeVar("_E", bound=Enum)


class EnumParameter(Parameter, Generic[_E]):
    """A parameter which takes the value of an Enum."""

    def __init__(
        self, enum: type[_E], /, *args: Any, case_sensitive: bool = True, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self._enum = enum
        self._case_sensitive = case_sensitive

    @override
    def normalize(self, x: _E | str) -> _E:
        return ensure_enum(self._enum, x, case_sensitive=self._case_sensitive)

    @override
    def parse(self, x: str) -> _E:
        return parse_enum(self._enum, x, case_sensitive=self._case_sensitive)

    @override
    def serialize(self, x: _E) -> str:
        return x.name


class TimeParameter(Parameter):
    """A parameter which takes the value of a `dt.time`."""

    @override
    def normalize(self, x: dt.time | str) -> dt.time:
        return ensure_time(x)

    @override
    def parse(self, x: str) -> dt.time:
        return parse_time(x)

    @override
    def serialize(self, x: dt.time) -> str:
        return serialize_time(x)


class WeekdayParameter(Parameter):
    """A parameter which takes the valeu of the previous/next weekday."""

    def __init__(
        self, *args: Any, rounding: Literal["prev", "next"] = "prev", **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        if rounding == "prev":
            self._rounder = round_to_prev_weekday
        else:
            self._rounder = round_to_next_weekday

    @override
    def normalize(self, x: dt.date | str) -> dt.date:
        with suppress(AttributeError, ModuleNotFoundError):
            from utilities.pandas import timestamp_to_date

            x = timestamp_to_date(x)
        return self._rounder(ensure_date(x))

    @override
    def parse(self, x: str) -> dt.date:
        return parse_date(x)

    @override
    def serialize(self, x: dt.date) -> str:
        return serialize_date(x)


__all__ = [
    "DateHourParameter",
    "DateMinuteParameter",
    "DateParameter",
    "DateSecondParameter",
    "EnumParameter",
    "TimeParameter",
    "WeekdayParameter",
]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from utilities.typed_settings import click_field


@dataclass(frozen=True)
class Config:
    """Settings for the `monitor_memory` script."""

    path: Path = click_field(default=Path("input.csv"), param_decls=("-p", "--path"))



from __future__ import annotations

import datetime as dt
from collections.abc import Callable
from operator import eq
from typing import Any

from hypothesis import assume, given
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    booleans,
    data,
    floats,
    integers,
    lists,
    none,
    sets,
)
from numpy import int64, isnan
from pandas import DataFrame
from pandas.testing import assert_frame_equal
from pytest import mark, param, raises
from sqlalchemy import (
    DECIMAL,
    Boolean,
    Column,
    Date,
    DateTime,
    Engine,
    Float,
    Integer,
    LargeBinary,
    MetaData,
    String,
    Table,
    select,
)

from utilities._sqlalchemy.pandas import (
    InsertPandasDataFrameError,
    StreamDataFramesError,
    TableColumnToDTypeError,
    _dataframe_columns_to_snake,
    _rows_to_dataframe,
    stream_dataframes,
    table_column_to_dtype,
)
from utilities.datetime import date_to_datetime
from utilities.hypothesis import dates_pd, datetimes_pd, sqlite_engines, text_ascii
from utilities.numpy import dt64ns
from utilities.pandas import Int64, boolean, check_dataframe, datetime64nsutc, string
from utilities.sqlalchemy import (
    ensure_tables_created,
    insert_items,
    insert_pandas_dataframe,
    select_to_pandas_dataframe,
)
from utilities.text import snake_case


class TestDataFrameColumnsToSnake:
    @given(col_name=text_ascii())
    def test_main(self, *, col_name: str) -> None:
        df = DataFrame(columns=[col_name])
        snake = _dataframe_columns_to_snake(df)
        assert snake.columns.tolist() == [snake_case(col_name)]


def _check_datetime(x: dt.datetime | None, y: dt.datetime | None, /) -> bool:
    return (x == y) or (
        isinstance(x, dt.datetime)
        and isinstance(y, dt.datetime)
        and (x.replace(tzinfo=None) == y.replace(tzinfo=None))
    )


def _check_float(x: float | None, y: float | None, /) -> bool:
    return (
        (x == y)
        or ((x is None) and isinstance(y, float) and isnan(y))
        or (isinstance(x, float) and isnan(x) and (y is None))
    )


class TestInsertPandasDataFrame:
    @given(data=data(), engine=sqlite_engines())
    @mark.parametrize(
        ("strategy", "pd_dtype", "col_type", "check"),
        [
            param(booleans(), bool, Boolean, eq),
            param(booleans() | none(), boolean, Boolean, eq),
            param(dates_pd() | none(), dt64ns, Date, eq),
            param(
                datetimes_pd().map(lambda x: x.replace(tzinfo=None)) | none(),
                dt64ns,
                DateTime,
                _check_datetime,
            ),
            param(
                datetimes_pd() | none(),
                datetime64nsutc,
                DateTime(timezone=True),
                _check_datetime,
            ),
            param(floats(), float, Float, _check_float),
            param(integers(-10, 10), int, Integer, eq),
            param(integers(-10, 10), int64, Integer, eq),
            param(integers(-10, 10) | none(), Int64, Integer, eq),
            param(text_ascii() | none(), string, String, eq),
        ],
    )
    def test_main(
        self,
        *,
        data: DataObject,
        engine: Engine,
        strategy: SearchStrategy[Any],
        pd_dtype: Any,
        col_type: Any,
        check: Callable[[Any, Any], bool],
    ) -> None:
        values = data.draw(lists(strategy))
        df = DataFrame(values, columns=["value"], dtype=pd_dtype)
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("value", col_type),
        )
        ensure_tables_created(engine, table)
        insert_pandas_dataframe(engine, df, table)
        sel = select(table.c["value"])
        with engine.begin() as conn:
            res = conn.execute(sel).scalars().all()
        for r, v in zip(res, values, strict=True):
            assert check(r, v)

    @given(
        engine=sqlite_engines(),
        date=dates_pd(),
        datetime=datetimes_pd().map(lambda x: x.replace(tzinfo=None)),
    )
    def test_mixed_date_and_datetimes(
        self, *, engine: Engine, date: dt.date, datetime: dt.datetime
    ) -> None:
        _ = assume(
            (datetime.hour != 0) or (datetime.minute != 0) or (datetime.second != 0)
        )
        df = DataFrame([date, datetime], columns=["value"], dtype=dt64ns)
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("value", DateTime),
        )
        ensure_tables_created(engine, table)
        insert_pandas_dataframe(engine, df, table)
        sel = select(table.c["value"])
        with engine.begin() as conn:
            res = conn.execute(sel).scalars().all()
        expected = [date_to_datetime(date, tzinfo=None), datetime]
        assert res == expected

    @given(engine=sqlite_engines(), values=lists(booleans() | none(), min_size=1))
    def test_error(self, *, engine: Engine, values: list[bool | None]) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("value", Boolean),
        )
        df = DataFrame(values, columns=["other"], dtype=boolean)
        with raises(InsertPandasDataFrameError):
            insert_pandas_dataframe(engine, df, table)


class TestRowsToDataFrame:
    @given(engine=sqlite_engines(), ids=sets(integers(0, 10)))
    @mark.parametrize(("col_name", "snake"), [param("id", False), param("Id", True)])
    def test_main(
        self, *, col_name: str, ids: set[int], engine: Engine, snake: bool
    ) -> None:
        table = Table(
            "example", MetaData(), Column(col_name, Integer, primary_key=True)
        )
        ensure_tables_created(engine, table)
        insert_items(engine, ([(id_,) for id_ in ids], table))
        with engine.begin() as conn:
            rows = conn.execute(sel := select(table)).all()
        df = _rows_to_dataframe(sel, rows, snake=snake)
        assert len(df) == len(ids)
        assert dict(df.dtypes) == {"id": Int64}


class TestSelectToPandasDataFrame:
    @given(data=data(), engine=sqlite_engines())
    @mark.parametrize(
        ("strategy", "pd_dtype_in", "pd_dtype_out", "col_type"),
        [
            param(booleans(), bool, boolean, Boolean),
            param(booleans() | none(), boolean, boolean, Boolean),
            param(dates_pd() | none(), dt64ns, dt64ns, Date),
            param(
                datetimes_pd().map(lambda x: x.replace(tzinfo=None)) | none(),
                dt64ns,
                dt64ns,
                DateTime,
            ),
            param(
                datetimes_pd() | none(),
                datetime64nsutc,
                datetime64nsutc,
                DateTime(timezone=True),
            ),
            param(floats(), float, float, Float),
            param(integers(-10, 10), int, Int64, Integer),
            param(integers(-10, 10), int64, Int64, Integer),
            param(integers(-10, 10) | none(), Int64, Int64, Integer),
            param(text_ascii() | none(), string, string, String),
        ],
    )
    def test_main(
        self,
        *,
        data: DataObject,
        engine: Engine,
        pd_dtype_in: Any,
        pd_dtype_out: Any,
        col_type: Any,
        strategy: SearchStrategy[Any],
    ) -> None:
        values = data.draw(lists(strategy))
        df = DataFrame(values, columns=["value"], dtype=pd_dtype_in)
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("value", col_type),
        )
        ensure_tables_created(engine, table)
        insert_pandas_dataframe(engine, df, table)
        sel = select(table.c["value"])
        result = select_to_pandas_dataframe(sel, engine)
        expected = DataFrame(values, columns=["value"], dtype=pd_dtype_out)
        assert_frame_equal(result, expected)

    @given(
        engine=sqlite_engines(),
        values=lists(booleans() | none()),
        stream=integers(1, 10),
    )
    def test_stream(
        self, *, engine: Engine, values: list[bool | None], stream: int
    ) -> None:
        df = DataFrame(values, columns=["value"], dtype=boolean)
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("value", Boolean),
        )
        ensure_tables_created(engine, table)
        insert_pandas_dataframe(engine, df, table)
        sel = select(table.c["value"])
        dfs = select_to_pandas_dataframe(sel, engine, stream=stream)
        for df in dfs:
            check_dataframe(
                df, dtypes={"value": boolean}, min_length=1, max_length=stream
            )

    @given(engine=sqlite_engines(), values=lists(booleans() | none()))
    def test_snake(self, *, engine: Engine, values: list[bool | None]) -> None:
        df = DataFrame(values, columns=["Value"], dtype=boolean)
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("Value", Boolean),
        )
        ensure_tables_created(engine, table)
        insert_pandas_dataframe(engine, df, table)
        sel = select(table.c["Value"])
        result = select_to_pandas_dataframe(sel, engine, snake=True)
        expected = DataFrame(values, columns=["value"], dtype=boolean)
        assert_frame_equal(result, expected)


class TestStreamDataFrames:
    @given(
        engine=sqlite_engines(),
        ids=sets(integers(min_value=0, max_value=10), min_size=1, max_size=10),
        stream=integers(1, 10),
    )
    def test_main(self, *, engine: Engine, ids: set[int], stream: int) -> None:
        table = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        ensure_tables_created(engine, table)
        insert_items(engine, ([(id_,) for id_ in ids], table))
        for df in stream_dataframes(select(table), engine, stream):
            assert 1 <= len(df) <= stream
            assert dict(df.dtypes) == {"id": Int64}

    @given(engine=sqlite_engines())
    def test_error(self, *, engine: Engine) -> None:
        table = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        with raises(StreamDataFramesError):
            _ = list(stream_dataframes(select(table), engine, 0))


class TestTableColumnToDType:
    @mark.parametrize(
        ("column", "expected"),
        [
            param(Column(Boolean), boolean),
            param(Column(Date), dt64ns),
            param(Column(DateTime), dt64ns),
            param(Column(DECIMAL), float),
            param(Column(Float), float),
            param(Column(Integer), Int64),
            param(Column(String), string),
        ],
    )
    def test_main(self, *, column: Any, expected: Any) -> None:
        assert table_column_to_dtype(column) == expected

    def test_error(self) -> None:
        column = Column("value", LargeBinary)
        with raises(TableColumnToDTypeError):
            _ = table_column_to_dtype(column)
